from __future__ import unicode_literals

import json
import re
import time
from lxml import etree

from ansible.errors import AnsibleError


class IPv4( object ):
    _ipv4_reg = re.compile( r"^([0-9]+)\.([0-9]+)\.([0-9]+)\.([0-9]+)(?:/([0-9]+))?$" )

    @classmethod
    def parse(cls, address ):
        m = cls._ipv4_reg.match( address )
        if m is None:
            raise ValueError("Value is not in IPv4 dotted format")

        octets = [ int(g) for g in m.group(1, 2, 3, 4) ]
        if m.group(5) is None:
            return (octets, 32)
        else:
            return (octets, int(m.group(5)))

    def __init__(self, address):
        self._octets, self._prefix_length = self.parse(address)

    @property
    def octets(self):
        return self._octets[:]

    @property
    def prefix_length(self):
        return self._prefix_length

    @property
    def CIDR(self):
        return "{}.{}.{}.{}/{pref}".format( *self.octets, pref=self.prefix_length )
    
    @property
    def address(self):
        return "{}.{}.{}.{}".format( *self.octets )

    def reverse_pointer_octets(self, reverse=True, trim=False):
        # https://tools.ietf.org/html/rfc1101
        # round mask down to octet as octet index
        mask = min((self._prefix_length-1) // 8, 3)
        # zero out host part of the network and reverse
        ip = list(self._octets)
        if trim:
            octets = [ o for i, o in enumerate(ip) if i <= mask ]
        else:
            octets = [ ( o if i <= mask else 0 ) for i,o in enumerate(ip) ]
        if reverse:
            octets.reverse()
        return octets
    
    @property
    def reverse_pointer(self):
        return "{}.{}.{}.{}.in-addr.arpa.".format(self.reverse_pointer_octets())

    @property
    def reverse_zone(self):
        octets = self.reverse_pointer_octets(reverse=True, trim=True)
        fmt = "{}." * len(octets) + "in-addr.arpa."
        return fmt.format(*octets)
    
    def __repr__(self):
        return "<IPv4 {}>".format(str(self))
    
    def __str__(self):
        return "{}.{}.{}.{}/{pref}".format(*self.octets, pref=self.prefix_length)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.CIDR == other.CIDR


class FilterModule(object):
    def __init__(self):
        self.ns_map = {
            'vnet': 'http://www.terhaak.de/~jojo/xml/vnet',
        }
    
    def filters(self):
        return {
            'ovs_json_table': self.ovs_json_table,
            'vnet_revdns_zone': self.vnet_revdns_zone,
            'vnet_time': self.vnet_time,
            'vnet_virt_swmap': self.vnet_virt_swmap,
        }

    def vnet_revdns_zone(self, input):
        return IPv4(input).reverse_zone

    def vnet_time(self, input):
        return int(time.time())
    
    def ovs_json_table(self, input):
        input = json.loads(input)
        head, data = (input["headings"], input["data"])
        output = [ {} for _ in  range(len(data)) ]
        
        for r in range(len(data)):
            for c in range(len(head)):
                cell = data[r][c]
                if type(cell) is list and cell[0] == "map":
                    cell = dict(cell[1])
                elif type(cell) is list and cell[0] == "set":
                    cell = list(cell[1])
                elif type(cell) is list and cell[0] == "uuid":
                    cell = cell[1]
                output[r][head[c]] = cell
        
        return output

    def _query(self, elem, path):
        res = elem.xpath(path, namespaces=self.ns_map)
        if len(res) > 0:
            return res[0].text
        return None

    def vnet_virt_swmap(self, value):
        swmap = {}
        for xmlstr in value:
            root = etree.fromstring(xmlstr)

            switch = self._query(root, "//network/metadata/vnet:switch/vnet:name")
            if switch is None:
                continue
            uuid = self._query(root, "//network/uuid")
            
            swmap[switch] = uuid

        return swmap
