import re
import collections

class Record:

    @staticmethod
    def make_name(zone_name, record_name):
        zone_name = zone_name.lower().strip('.')
        record_name = record_name.lower().strip('.')
        
        if record_name.endswith('.' + zone_name):
            return record_name
        return record_name + '.' + zone_name
    
    def __init__(self, r_id, domain_id, name, type, content, ttl):
        self.id = r_id
        self.domain_id = domain_id
        self.name = name
        self.type = type.upper()
        self.content = content
        self.ttl = ttl
    
    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        tup = lambda x: x.id, x.domain_id, x.name, x.type, x.content, x.ttl
        return tup(self) == tup(other)

    def same_as(self, other):
        tup = lambda x: (x.domain_id, x.name, x.type, x.content, x.ttl)
        return tup(self) == tup(other)


class SoaRecord(Record):
    
    @classmethod
    def default_for(cls, zone):
        pri, mail = ("a.misconfigured.dns.server.invalid", "hostmaster.gaga.test")
        ser, refr, retr, exp, nttl = (0, 10800, 3600, 604800, 3600)
        data = (pri, mail, ser, refr, retr, exp, nttl)
        return cls(None, zone.id, zone.name, 'SOA', data, 3600)
    
    def __init__(self, r_id, domain_id, name, type, content, ttl):
        super().__init__(r_id, domain_id, name, 'SOA', '', ttl)
        
        if isinstance(content, tuple):
            self._content = list(content)
        else:
            self._content = list(re.split('\s+', content))

        if len(self._content) != 7:
            raise ValueError("invalid SOA record length")
        
    @property
    def content(self):
        return "{} {} {} {} {} {} {}".format(*self._content)

    @content.setter
    def content(self, value):
        pass

    @property
    def serial(self):
        return int(self._content[2])
    
    @serial.setter
    def serial(self, value):
        self._content[2] = value
    
    def increment_serial(self):
        self.serial = self.serial + 1
    

class Zone:
    def __init__(self, zone_id, name, repl_type, soa=None, meta=None):
        self.id = zone_id
        self.name = name
        self.type = repl_type

        if soa is None:
            soa = SoaRecord.default_for(self)
        self.soa = soa

        if meta is None:
            meta = MetadataDict()
        self.metadata = meta

    def increment_serial(self):
        self.soa.increment_serial()


class MetadataDict(collections.UserDict):
    def put(self, key, value):
        if key in self:
            self[key].append(str(value))
        else:
            self[key] = str(value)
    
    def iter_values(self):
        for key, values in self.items():
            for v in values:
                yield (key, v)
    
    def __setitem__(self, key, value):
        if isinstance(value, (list, tuple, set, frozenset)):
            value = [str(v) for v in value]
        else:
            value = [str(value)]
        super().__setitem__(str(key).upper(), value)

    def __getitem__(self, key):
        return super().__getitem__(str(key).upper())


class RecordRepo:
    def __init__(self, con):
        self.con = con
        self.changed = False

    def create(self, record):
        q =  "insert into records (domain_id, name, type, content, ttl) values (?, ?, ?, ?, ?)"
        data = [record.domain_id, record.name, record.type, record.content, record.ttl]
        cur = self.con.execute(q, data)
        record.id = cur.lastrowid
        self.changed = True

    def find(self, record, ignore_content=False):
        q = "select id, domain_id, name, type, content, ttl from records where"
        if record.id:
            q += " id = ?"
            data = [record.id]
        else:
            q += " domain_id = ? and name = ? and type = ?"
            data = [record.domain_id, record.name, record.type]
            if not ignore_content:
                q += " and content = ?"
                data += [record.content]
        
        cur = self.con.execute(q, data)
        return [Record(r[0], r[1], r[2], r[3], r[4], r[5]) for r in cur]

    def list(self, zone_id=None, name=None):
        q, data = ("select id, domain_id, name, type, content, ttl from records", [])
        fields = [f for f in (('domain_id', zone_id), ('name', name)) if f[1]]
        if fields:
            q += " where " + ' and '.join([f[0] + " = ?" for f in fields])
            data += [f[1] for f in fields]
        cur = self.con.execute(q, data)
        return [Record(r[0], r[1], r[2], r[3], r[4], r[5]) for r in cur]

    def get(self, record, **kwargs):
        records = self.find(record, **kwargs)
        return records[0] if records else None

    def save(self, record):
        old = self.get(record, ignore_content=False)
        if old:
            fields = []
            if old.content != record.content:
                fields.append(("content", record.content))
            if old.ttl != record.ttl:
                fields.append(("ttl", record.ttl))
            
            if fields:
                q =  "update records set " 
                q += ", ".join(["{} = ?".format(f[0]) for f in fields])
                q += "where id = ?"
                data = [f[1] for f in fields] + [old.id]
                self.con.execute(q, data)
                self.changed = True

    def delete(self, record):
        old = self.get(record, ignore_content=False)
        if old:
            self.con.execute("delete from records where id = ?", [old.id])
            self.changed = True


class MetadataRepo:
    def __init__(self, con):
        self.con = con
        self.changed = False
    
    def find_for(self, zone_id):
        q = "select kind, content from domainmetadata where domain_id = ?"
        meta = MetadataDict()
        for row in self.con.execute(q, [zone_id]):
            meta.put(row[0], row[1])
        return meta
    
    def save(self, zone_id, metadata):
        q = "select id, kind, content from domainmetadata where domain_id = ?"
        rows, index = (list(self.con.execute(q, [zone_id])), set())

        q = "delete from domainmetadata where id = ?"
        for row in rows:
            index.add((row[1], row[2]))
            if row[2] not in metadata.get(row[1], []):
                self.con.execute(q, [row[0]])
                self.changed = True

        q = "insert into domainmetadata (domain_id, kind, content) values (?, ?, ?)"
        for key, value in metadata.iter_values():
            if (key, value) not in index:
                self.con.execute(q, [zone_id, key, value])
                self.changed = True
    

class ZoneRepo:
    def __init__(self, con, metadata=None):
        self.con = con
        self.changed = False
        
        if metadata is None:
            metadata = MetadataRepo(self.con)
        self.metadata = metadata
        
        con.execute("PRAGMA foreign_keys = ON")

    def create(self, zone):
        q = "select id from domains where name = ?"
        cur = self.con.execute(q, [zone.name])
        if cur.fetchone() is not None:
            return
        
        q = "insert into domains (name, type) values (?, ?)"
        cur = self.con.execute(q, [zone.name, zone.type])
        zone.id = cur.lastrowid
        zone.soa.domain_id = zone.id
        self.changed = True
        
        q =  "insert into records (domain_id, name, type, content, ttl) values (?, ?, ?, ?, ?)"
        data = [zone.soa.domain_id, zone.name, 'SOA', zone.soa.content, zone.soa.ttl]
        cur = self.con.execute(q, data)
        zone.soa.id = cur.lastrowid
        self.changed = True
        
        self.metadata.save(zone.id, zone.metadata)
    
    def find(self, zone):
        q = """
        select 
            domains.id, domains.name, domains.type,
            records.id as soa_id, records.content as soa_cont, records.ttl as soa_ttl
        from domains
        left join records on 
            records.domain_id = domains.id and records.type = 'SOA'
        where {field} = ?
        """

        if zone.id:
            field, value = ("domains.id", zone.id)
        else:
            field, value = ("domains.name", zone.name)

        cur = self.con.execute(q.format(field=field), [value])
        zones = []
        for r in cur:
            soa = SoaRecord(r[3], r[0], r[1], 'SOA', r[4], r[5])
            meta = self.metadata.find_for(r[0])
            zones.append(Zone(r[0], r[1], r[2], soa=soa, meta=meta))
        return zones

    def list(self):
        q = """
        select 
            domains.id, domains.name, domains.type,
            records.id as soa_id, records.content as soa_cont, records.ttl as soa_ttl
        from domains
        left join records on 
            records.domain_id = domains.id and records.type = 'SOA'
        """
        zones = []
        for r in self.con.execute(q):
            soa = SoaRecord(r[3], r[0], r[1], 'SOA', r[4], r[5])
            meta = self.metadata.find_for(r[0])
            zones.append(Zone(r[0], r[1], r[2], soa=soa, meta=meta))
        return zones

    def get(self, zone):
        zones = self.find(zone)
        return zones[0] if zones else None

    def get_by_name(self, zone_name):
        return self.get(Zone(None, zone_name, 'NATIVE'))
    
    def save(self, zone):
        old = self.get(zone)
        if not old:
            return
        
        if old.type != zone.type:
            q =  "update domains set type = ? where id = ?" 
            self.con.execute(q, [old.type, old.id])
            self.changed = True
        
        if not old.soa.same_as(zone.soa) and old.soa.id:
            q = "update records set content = ?, ttl = ? where id = ?"
            self.con.execute(q, [zone.soa.content, zone.soa.ttl, zone.soa.id])
            self.changed = True

        self.metadata.save(zone.id, zone.metadata)

    def delete(self, zone):
        old = self.get(zone)
        if old:
            self.con.execute("delete from domains where id = ?", [old.id])
            self.changed = True
